//  Boost exact.hpp header file (proposed) -----------------------------------//

//  (C) Copyright Scott McMurray 2006

//  Use, modification, and distribution is subject to the Boost Software
//  License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy
//  at http://www.boost.org/LICENSE_1_0.txt)

//  See library home page at http://www.boost.org/libs/???????

//----------------------------------------------------------------------------//

#ifndef BOOST_EXACT_HPP
#define BOOST_EXACT_HPP

#include <boost/detail/endian.hpp>
//#include <boost/cast.hpp>
#include <boost/type_traits/is_signed.hpp>
#include <boost/static_assert.hpp>
#include <boost/integer.hpp>
#include <limits>
#include <cstring>
//#include <boost/integer_cover_operators.hpp>
#include "integer_cover_operators.hpp"

namespace boost {

template <typename T>
struct wrapped_type {
    typedef T type;
};

namespace detail {

enum { bits_per_byte = std::numeric_limits<unsigned char>::digits };

template <int x, int y>
struct round_up_div {
    enum { value = (x+y-1)/y };
};

template <bool is_signed, size_t n_bits>
struct native_type_impl;

template <size_t n_bits>
struct native_type_impl<true,n_bits> {
    typedef typename boost::int_t<n_bits>::fast type;
};
template <size_t n_bits>
struct native_type_impl<false,n_bits> {
    typedef typename boost::uint_t<n_bits>::fast type;
};

// So that output shows as integers, not characters
template<>
struct native_type_impl<true, bits_per_byte> {
    typedef signed type;
};
template<>
struct native_type_impl<false, bits_per_byte> {
    typedef unsigned type;
};


template <typename S, size_t n_bits>
struct native_type {
    typedef typename native_type_impl< boost::is_signed<S>::value,
                                       n_bits >::type type;
};

template <size_t n_bits>
struct align_type {
    typedef typename boost::uint_t<n_bits>::least type;
    BOOST_STATIC_ASSERT( n_bits == std::numeric_limits<type>::digits );
};

} // namespace detail

template < size_t n_bits,
           typename S = signed,
           size_t align_bits = detail::bits_per_byte >
class exact
 : boost::integer_cover_operators< exact<n_bits,S,align_bits>,
                                   typename detail::native_type<S,n_bits>::type > {
    BOOST_STATIC_ASSERT( n_bits % detail::bits_per_byte == 0 );
    typedef typename detail::native_type<S,n_bits>::type native_type;
    typedef typename detail::align_type<align_bits>::type align_type;
    enum { n_bytes = n_bits / detail::bits_per_byte };
    enum { native_bytes = sizeof(native_type) };
  public:
    typedef native_type value_type;
    enum { is_signed = boost::is_signed<S>::value };
    exact() {}
    exact(native_type i) {
        #ifdef BOOST_BIG_ENDIAN
            char * const icp = reinterpret_cast<char*>( &i );
            std::memcpy( value, icp + (native_bytes-n_bytes), n_bytes );
        #else
            std::memcpy( value, &i, n_bytes );
        #endif
    }
    operator native_type() const {
        native_type i = 0;
        #ifdef BOOST_BIG_ENDIAN
            char * const icp = reinterpret_cast<char*>( &i );
            std::memcpy( icp + (native_bytes-n_bytes), value, n_bytes );
        #else
            std::memcpy( &i, value, n_bytes );
        #endif
        if ( is_signed && n_bytes != sizeof(native_type) ) {
            native_type const mask = native_type(1) << ( n_bits - 1 );
            return -(i & mask) | i;
        }
        return i;
    }
  private:
    union {
        char value[ n_bytes ];
        align_type alignment;
    };
};

template < size_t n_bits,
           typename S,
           size_t align_bits >
struct is_signed< exact<n_bits,S,align_bits> >
 : boost::is_signed<S>::value {};

template < size_t n_bits,
           typename S,
           size_t align_bits >
struct wrapped_type< exact<n_bits, S, align_bits> > {
    typedef typename exact<n_bits, S, align_bits>::value_type type;
};


} // namespace boost

namespace std {

template < size_t n_bits,
           typename S,
           size_t align_bits >
struct numeric_limits< boost::exact<n_bits,S,align_bits> > 
 : numeric_limits< typename boost::exact<n_bits,S,align_bits>::value_type > {

    typedef boost::exact<n_bits,S,align_bits> this_type;
    typedef typename this_type::value_type value_type;

    BOOST_STATIC_CONSTANT( bool, is_specialized = true );

    BOOST_STATIC_CONSTANT( bool, is_signed = this_type::is_signed );
    BOOST_STATIC_CONSTANT( int, digits = n_bits - is_signed );

    // this magic division is an approximation of ceil( digits*log(2)/log(10) )
    BOOST_STATIC_CONSTANT( int, digits10 = (boost::detail
                                                 ::round_up_div< digits*301,
                                                                 1000
                                                               >::value) );

    static this_type max() {
        if ( this_type::is_signed ) {
            // digits doesn't include sign bit, so this shifts into sign bit
            return ( value_type(1) << digits ) - 1;
        } else {
            return ( ( value_type(1) << (digits-1) ) - 1 )
                 | ( value_type(1) << (digits-1) );
        }
    }
    static this_type min() { 
        if ( this_type::is_signed ) {
            return ~max();//value_type(1) << (digits-1);
        } else {
            return 0;        
        }
    }
};

} // namespace std


#endif // BOOST_EXACT_HPP
